#  Copyright © 2025 CloudBlue. All rights reserved.

import logging

from django.db import models, transaction
from django.dispatch import Signal
from django.utils.timezone import now

from dj_cqrs.constants import SignalType
from dj_cqrs.controller import producer
from dj_cqrs.dataclasses import TransportPayload
from dj_cqrs.state import cqrs_state
from dj_cqrs.utils import get_message_expiration_dt


logger = logging.getLogger('django-cqrs')

post_bulk_create = Signal()
"""
Signal sent after a bulk create.
See dj_cqrs.mixins.RawMasterMixin.call_post_bulk_create.
"""

post_update = Signal()
"""
Signal sent after a bulk update.
See dj_cqrs.mixins.RawMasterMixin.call_post_update.
"""


class MasterSignals:
    """Signals registry and handlers for CQRS master models."""

    @classmethod
    def register_model(cls, model_cls):
        """
        Registers signals for a model.

        Args:
            model_cls (dj_cqrs.mixins.MasterMixin): Model class inherited from CQRS MasterMixin.
        """

        models.signals.post_save.connect(cls.post_save, sender=model_cls)
        models.signals.post_delete.connect(cls.post_delete, sender=model_cls)

        post_bulk_create.connect(cls.post_bulk_create, sender=model_cls)
        post_update.connect(cls.post_bulk_update, sender=model_cls)

    @classmethod
    def post_save(cls, sender, **kwargs):
        """
        Args:
            sender (dj_cqrs.mixins.MasterMixin): Class or instance inherited from CQRS MasterMixin.
        """
        if not sender.CQRS_PRODUCE:
            return

        update_fields = kwargs.get('update_fields')
        if update_fields and ('cqrs_revision' not in update_fields):
            return

        instance = kwargs['instance']
        if not instance.is_sync_instance():
            return

        using = kwargs['using']

        bulk_relate_cm = getattr(cqrs_state, 'bulk_relate_cm', None)
        if bulk_relate_cm:
            bulk_relate_cm.register(instance, using)

        sync = kwargs.get('sync', False)
        queue = kwargs.get('queue', None)

        connection = transaction.get_connection(using)
        if not connection.in_atomic_block or instance.is_initial_cqrs_save:
            transaction.on_commit(
                lambda: cls._post_save_produce(sender, instance, using, sync, queue),
            )

    @classmethod
    def _post_save_produce(cls, sender, instance, using, sync, queue):
        # As this method may run 'on_commit', the instance may not exist. In that case, log the
        # error but don't raise an exception.
        try:
            instance.reset_cqrs_saves_count()
            instance_data = instance.to_cqrs_dict(using, sync=sync)
            previous_data = instance.get_tracked_fields_data()
            signal_type = SignalType.SYNC if sync else SignalType.SAVE
            meta = instance.get_cqrs_meta(
                instance_data=instance_data,
                previous_data=previous_data,
                signal_type=signal_type,
            )
        except sender.DoesNotExist:
            logger.error(
                f"Can't produce message from master model '{sender.__name__}': "
                f"The instance doesn't exist (pk={instance.pk})",
            )
            return

        payload = TransportPayload(
            signal_type,
            sender.CQRS_ID,
            instance_data,
            instance.pk,
            queue,
            previous_data,
            expires=get_message_expiration_dt(),
            meta=meta,
        )
        producer.produce(payload)

    @classmethod
    def post_delete(cls, sender, **kwargs):
        """
        Args:
            sender (dj_cqrs.mixins.MasterMixin): Class or instance inherited from CQRS MasterMixin.
        """
        if not sender.CQRS_PRODUCE:
            return

        instance = kwargs['instance']
        if not instance.is_sync_instance():
            return

        instance_data = {
            'id': instance.pk,
            'cqrs_revision': instance.cqrs_revision + 1,
            'cqrs_updated': str(now()),
        }

        data = instance.get_custom_cqrs_delete_data()
        if data:
            instance_data['custom'] = data

        signal_type = SignalType.DELETE

        meta = instance.get_cqrs_meta(
            instance_data=instance_data,
            signal_type=signal_type,
        )

        payload = TransportPayload(
            signal_type,
            sender.CQRS_ID,
            instance_data,
            instance.pk,
            expires=get_message_expiration_dt(),
            meta=meta,
        )
        # Delete is always in transaction!
        transaction.on_commit(lambda: producer.produce(payload))

    @classmethod
    def post_bulk_create(cls, sender, **kwargs):
        """
        Args:
            sender (dj_cqrs.mixins.MasterMixin): Class or instance inherited from CQRS MasterMixin.
        """
        cls._post_bulk(sender, **kwargs)

    @classmethod
    def post_bulk_update(cls, sender, **kwargs):
        """
        Args:
            sender (dj_cqrs.mixins.MasterMixin): Class or instance inherited from CQRS MasterMixin.
        """
        cls._post_bulk(sender, **kwargs)

    @classmethod
    def _post_bulk(cls, sender, **kwargs):
        for instance in kwargs['instances']:
            cls.post_save(sender, instance=instance, using=kwargs['using'])



import logging

from django.conf import settings
from django.db import OperationalError, transaction

from dj_cqrs.constants import (
    DB_VENDOR_MYSQL,
    DB_VENDOR_PG,
    MYSQL_TIMEOUT_ERROR_CODE,
    PG_TIMEOUT_FLAG,
    SUPPORTED_TIMEOUT_DB_VENDORS,
)


def install_last_query_capturer(model_cls):
    conn = _connection(model_cls)
    if not _get_last_query_capturer(conn):
        conn.execute_wrappers.append(_LastQueryCaptureWrapper())


def log_timed_out_queries(error, model_cls):  # pragma: no cover
    log_q = bool(settings.CQRS['replica'].get('CQRS_LOG_TIMED_OUT_QUERIES', False))
    if not (log_q and isinstance(error, OperationalError) and error.args):
        return

    conn = _connection(model_cls)
    conn_vendor = getattr(conn, 'vendor', '')
    if conn_vendor not in SUPPORTED_TIMEOUT_DB_VENDORS:
        return

    e_arg = error.args[0]
    is_timeout_error = bool(
        (conn_vendor == DB_VENDOR_MYSQL and e_arg == MYSQL_TIMEOUT_ERROR_CODE)
        or (conn_vendor == DB_VENDOR_PG and isinstance(e_arg, str) and PG_TIMEOUT_FLAG in e_arg)
    )
    if is_timeout_error:
        query = getattr(_get_last_query_capturer(conn), 'query', None)
        if query:
            logger_name = settings.CQRS['replica'].get('CQRS_QUERY_LOGGER', '') or 'django-cqrs'
            logger = logging.getLogger(logger_name)
            logger.error('Timed out query:\n%s', query)


class _LastQueryCaptureWrapper:
    def __init__(self):
        self.query = None

    def __call__(self, execute, sql, params, many, context):
        try:
            execute(sql, params, many, context)
        finally:
            self.query = sql


def _get_last_query_capturer(conn):
    return next((w for w in conn.execute_wrappers if isinstance(w, _LastQueryCaptureWrapper)), None)


def _connection(model_cls):
    return transaction.get_connection(using=model_cls._default_manager.db)



#  Copyright © 2025 CloudBlue. All rights reserved.

import logging
from collections import defaultdict
from contextlib import ContextDecorator
from datetime import date, datetime, timedelta
from uuid import UUID

from django.conf import settings
from django.db import transaction
from django.utils import timezone

from dj_cqrs.constants import DB_VENDOR_PG, SUPPORTED_TIMEOUT_DB_VENDORS
from dj_cqrs.logger import install_last_query_capturer
from dj_cqrs.state import cqrs_state


logger = logging.getLogger('django-cqrs')


def get_message_expiration_dt(message_ttl=None):
    """Calculates when message should expire.

    :param int or None message_ttl:
    :return: Expiration datetime or None if infinite
    :rtype: datetime.datetime or None
    """
    message_ttl = message_ttl or settings.CQRS['master']['CQRS_MESSAGE_TTL']
    if message_ttl is None:
        # Infinite
        return

    return timezone.now() + timedelta(seconds=message_ttl)


def get_delay_queue_max_size():
    """Returns max allowed number of "waiting" messages in the delay queue.

    :return: Positive integer number or None if infinite
    :rtype: int
    """
    if 'replica' not in settings.CQRS:
        return None

    return settings.CQRS['replica']['delay_queue_max_size']


def get_messages_prefetch_count_per_worker():
    """Returns max allowed number of unacked messages, that can be consumed by a single worker.

    :return: Positive integer number or 0 if infinite
    :rtype: int
    """
    delay_queue_max_size = get_delay_queue_max_size()
    if delay_queue_max_size is None:
        # Infinite
        return 0

    return delay_queue_max_size + 1


def get_json_valid_value(value):
    return str(value) if isinstance(value, (date, datetime, UUID)) else value


def apply_query_timeouts(model_cls):  # pragma: no cover
    query_timeout = int(settings.CQRS['replica'].get('CQRS_QUERY_TIMEOUT', 0))
    if query_timeout <= 0:
        return

    model_db = model_cls._default_manager.db
    conn = transaction.get_connection(using=model_db)
    conn_vendor = getattr(conn, 'vendor', '')
    if conn_vendor not in SUPPORTED_TIMEOUT_DB_VENDORS:
        return

    if conn_vendor == DB_VENDOR_PG:
        statement = 'SET statement_timeout TO %s'
    else:
        statement = 'SET SESSION MAX_EXECUTION_TIME=%s'

    with conn.cursor() as cursor:
        cursor.execute(statement, params=(query_timeout,))

    install_last_query_capturer(model_cls)


class _BulkRelateCM(ContextDecorator):
    def __init__(self, cqrs_id=None):
        self._cqrs_id = cqrs_id
        self._mapping = defaultdict(lambda: defaultdict(set))
        self._cache = {}

    def register(self, instance, using=None):
        instance_cqrs_id = getattr(instance, 'CQRS_ID', None)
        if (not instance_cqrs_id) or (self._cqrs_id and instance_cqrs_id != self._cqrs_id):
            return

        self._mapping[instance_cqrs_id][using].add(instance.pk)

    def get_cached_instance(self, instance, using=None):
        instance_cqrs_id = getattr(instance, 'CQRS_ID', None)
        if (not instance_cqrs_id) or (self._cqrs_id and instance_cqrs_id != self._cqrs_id):
            return

        instance_pk = instance.pk
        cached_instances = self._cache.get(instance_cqrs_id, {}).get(using, {})
        if cached_instances:
            return cached_instances.get(instance_pk)

        cached_pks = self._mapping[instance_cqrs_id][using]
        if not cached_pks:
            return

        qs = instance.__class__._default_manager.using(using)
        instances_cache = {
            instance.pk: instance
            for instance in instance.__class__.relate_cqrs_serialization(qs)
            .filter(
                pk__in=cached_pks,
            )
            .order_by()
            .all()
        }
        self._cache.update(
            {
                instance_cqrs_id: {
                    using: instances_cache,
                },
            }
        )
        return instances_cache.get(instance_pk)

    def __enter__(self):
        cqrs_state.bulk_relate_cm = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        cqrs_state.bulk_relate_cm = None


def bulk_relate_cqrs_serialization(cqrs_id=None):
    return _BulkRelateCM(cqrs_id=cqrs_id)
